#!/usr/bin/perl -w

#
# Filter.pm
# $Id: Filter.pm 622 2008-01-26 05:46:57Z johnh $
#
# Copyright (C) 2007 by John Heidemann <johnh@isi.edu>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2, as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#


package Jdb::Filter;

=head1 NAME

Jdb::Filter - base class for Jdb filters

=head1 DESCRIPTION

Jdb::Filter is the virtual base class for Jdb filters.

Users will typically invoke individual programs via the command line
(for example, see L<dbcol(1)>)
or string together several in a Perl program
as described in L<dbpipeline(3)>.

For new Filter developers, internal processing is:

    new
	set_defaults
	parse_options
	autorun if desired
    parse_options   # optionally called additional times
    setup           # does IO on header
    run             # does IO on data
    finish          # any shutdown

In addition, the C<info> method returns metadata about a given filter.

=head1 FUNCTIONS

=cut

@ISA = ();
($VERSION) = ('$Revision: 622 $' =~ m/(\d+)/);

use strict;
use Carp qw(croak);
use IO::Handle;
use IO::File;
use Getopt::Long qw(:config no_ignore_case bundling permute);

use Jdb::IO::Reader;
use Jdb::IO::Writer;
use Jdb::Support;
use Jdb::Support::DelayPassComments;

=head2 new

    $jdb = new Jdb::Filter;

Create a new filter object, calling set_defaults
and parse_options.  A user program will call a specific
filter (say Jdb::Filter::dbcol) to do processing.
See also L<dbpipeline> for aliases that remove the wordiness.

=cut

sub new {
    my ($class) = @_;
    my $self = bless {}, $class;
    $self->set_defaults;
    return $self;
}

=head2 post_new

    $filter->post_new();

Called when the subclass is done with new,
giving Jdb::Filter a chance to autorun.

=cut

sub post_new {
    my $self = shift @_;
    $self->setup_run_finish()
        if ($self->{_autorun});
}


=head2 set_defaults

    $filter->set_defaults();

Set up object defaults.
Called once during new.

Jdb::Filter::set_defaults does some general setup,
tracking module invocation and preparing for one input and output stream.


=cut

sub set_defaults ($) {
    my $self = shift @_;

    # see the info() method for documentation
    $self->{_info} = {
	input_type => 'jdb*',
	input_count => 1,
	output_type => 'jdb*',
	output_count => 1,
    };

    my($package, $filename, $line) = caller(1);
    $filename =~ s@^.*/@@g;   # strip junk
    $filename =~ s@\.pm$@@;   # strip junk
    $self->{_prog} = $filename;
    $self->{_orig_argv} = undef;
    $self->{_logprog} = 1;
    $self->{_close} = 1;
    $self->{_empty} = '-';

    $self->{_debug} = undef;
    $self->{_error} = undef;

    # Jdb::IO objects
    $self->{_in} = undef;
    $self->{_out} = undef;
    # filehandles or Jdb::BoundedQueue:
    $self->{_input} = *STDIN;
    $self->{_output} = *STDOUT;

    $self->{_autorun} = undef;
};

=head2 parse_options

    $filter->parse_options(@ARGV);

Parse_options is called one I<or more> times to parse ARGV-style options.
It should not do any IO or any irreverssable actions; defer those to startup.

Jdb::Filter::parse_options does no work; the subclass is expected
to call Jdb::Filter::get_options() for all arguments.

Most modules implement certain common jdb options, listed below.

=for comment
(Note that this text in Jdb::Filter.pm is the master copy,
replicated in to all the module.pm files via C<make standardoptions>.)

=for comment
begin_standard_jdb_options

This module also supports the standard jdb options:

=over

=item B<-d>

Enable debugging output.

=item B<-i> or B<--input> InputSource

Read from InputSource, typically a file name, or C<-> for standard input,
or (if in Perl) a IO::Handle, Jdb::IO or Jdb::BoundedQueue objects.

=item B<-o> or B<--output> OutputDestination

Write to OutputDestination, typically a file name, or C<-> for standard output,
or (if in Perl) a IO::Handle, Jdb::IO or Jdb::BoundedQueue objects.

=item B<--autorun> or B<--noautorun>

By default, programs process automatically,
but Jdb::Filter objects in Perl do not run until you invoke
the run() method.
The C<--(no)autorun> option controls that behavior within Perl.

=item B<--noclose>

By default, programs close their output when done.
For some cases where objects are used internally, C<--noclose>
may be used to leave output open for further I/O.
(This option is only supported by some filters.)

=item B<--help>

Show help.

=item B<--man>

Show full manual.

=back

=for comment
end_standard_jdb_options


=cut

sub parse_options ($@) {
    my $self = shift @_;
    die "Jdb:Filter: we expect the subprogram to handle parse_options.\n";
}

=head2 get_options

    $success = $filter->get_options(\@argv, "v+" => \$verbose, ...)

get_options is just like Getopt::Long's GetOptions,
but takes the argument list as the first argument.
This list is modified and any non-options are returned.
It also saves _orig_argv in itself.

=cut

sub get_options ($$@) {
    my $self = shift @_;
    my $argv_ref = shift @_;
    my @opt_specs = @_;
    # hacky interface to GetOptions, we have to copy to and from ARGV
    my @saved_argv = @ARGV;   # paranoia
    foreach my $p (@$argv_ref) {
	my $p_copy = $p;
	if (ref($p_copy) =~ /\:\:/) {
	    # remove the internal memory cruft for internal perl objects
	    $p_copy = ref($p_copy);
	    $p_copy =~ s/\=.*$//;
	    $p_copy = "[$p_copy]";
        };
	push (@{$self->{_orig_argv}}, $p_copy);
    };
    @ARGV = @$argv_ref;
    my $result = GetOptions(@opt_specs);
    @$argv_ref = @ARGV;   # copy out unused arguments
    @ARGV = @saved_argv;
    return $result;
}

=head2 parse_sort_option

    $jdb_io = $filter->parse_sort_option($option_name, $target);

This helper function handles sorting options and column names
as described in L<dbsort(1)>.  We normalize long sort options
to unbundled short options
and accumulate them in $self->{_sort_argv}.

=cut

sub parse_sort_option ($$$) {
    my $self = shift @_;
    my ($option_name, $target) = @_;

    my $sort_aref = $self->{_sort_argv};
    if ($option_name eq '<>') {
	push (@$sort_aref, $target);
    } elsif ($option_name eq 'n' || $option_name eq 'numeric') {
	push (@$sort_aref, '-n');
    } elsif ($option_name eq 'N' || $option_name eq 'lexical') {
	push (@$sort_aref, '-N');
    } elsif ($option_name eq 'r' || $option_name eq 'descending') {
	push (@$sort_aref, '-r');
    } elsif ($option_name eq 'R' || $option_name eq 'descending') {
	push (@$sort_aref, '-R');
    } else {
	croak "parse_sort_option: unknown option $option_name for target $target\n";
    };
}



=head2 parse_io_option

    $jdb_io = $filter->parse_io_option($io_direction, $option_name, $target);

This helper function handles C<--input> or C<--output> options.
Parse an input or output option (using those strings in $IO_DIRECTION), 
for a $TARGET of either - for stdin/stdout, a filename,
or a Perl BoundedQueue object.
It fills in $self->{_$IO_DIRECTION} with the resulting object,
which is either afile handle or Jdb::Filter::Piepline object,
and expects finish_io_option to convert this token into a full Jdb::IO object.

$IO_DIRECTION is usually input or output, but it can also be inputs
(with an "s") when multiple input sources are allowed.

=cut

sub parse_io_option ($$$$) {
    my $self = shift @_;
    my $direction = shift @_;
    my $option_name = shift @_;
    my $target = shift @_;

    my($mode) = $direction eq 'output' ? "w" : "r";

    my $token;   # either a IO::Handle or IO::File or a Jdb::BoundedQueue or a Jdb::IO object
    if (!defined($target) || (ref($target) eq '' && $target eq '-')) {
	# special case the - string
	$token = new IO::Handle;
	$token->fdopen($direction =~ /^input/ ? fileno(STDIN) : fileno(STDOUT),
	    $mode) or croak "cannot open $direction ($mode) for IO.\n";
    } elsif (ref($target) eq '' && ($target eq '*main::STDIN' || $target eq '*main::STDOUT')) {
	$token = $target;
    } elsif (ref($target) eq '') {
	# string, so must be a filename
	$token = new IO::File $target, $mode;
	$token or croak "cannot open $target for $direction for IO.\n";
    } elsif (ref($target) =~ /^(Jdb::BoundedQueue|IO::|Jdb::IO)/) {
	# an approved existing perl object, so just pass it.
	$token = $target;
    } else {
	croak "cannot understand IO argument given to option $option_name in direction $direction\n";
    };

    if ($direction eq 'input' || $direction eq 'output') {
        $self->{"_$direction"} = $token;
    } elsif ($direction eq 'inputs' || $direction eq 'outputs') {
        push (@{$self->{"_$direction"}}, $token);
    } else {
	croak "internal error: bad direction";
    };
}

=head2 finish_io_option

    $jdb_io = $filter->finish_io_option($io_direction, @jdb_args);

This helper function 
finishes setting up a Jdb::IO object in $IO_DIRECTION,
using @JDB_ARGS as parameters.
It creates the actual Jdb::IO objects, opens the files (or whatever),
and reads the headers.
the resulting Jdb::IO objects are 
built from C<$self->{_$IO_DIRECTION}>
and are left in C<$self->{_in}> or (C<_out> or C<@_ins>).

$IO_DIRECTION must be "input", "inputs" or "output".

Since it does IO, finish_io_option should only be called from setup,
not parse_options.

Can be called once per IO stream.

=cut

sub finish_io_option ($$@) {
    my $self = shift @_;
    my $direction = shift @_;

    # special case multiple inputs
    if ($direction eq 'inputs') {
	my $token;
	foreach $token (@{$self->{_inputs}}) {
	    my $token_type = (ref($token) =~ /^Jdb::BoundedQueue/) ? "-queue" : "-fh";
	    my $in = new Jdb::IO::Reader($token_type => $token, @_);
	    $in->error and croak $self->{_prog} . ": cannot open input: " . $in->error . ".\n";
	    push @{$self->{_ins}}, $in;
        };
	return;
    };

    # single
    my $token = $self->{"_$direction"};
    my $jdb;
    if (ref($token) =~ /^Jdb::IO/) {
	# assume the user gave us a good one
	$jdb = $token;
    } else {
	my $token_type = (ref($token) =~ /^Jdb::BoundedQueue/) ? "-queue" : "-fh";
	if ($direction eq 'input') {
	    $jdb = new Jdb::IO::Reader($token_type => $token, @_);
	} elsif ($direction eq 'output') {
	    $jdb = new Jdb::IO::Writer($token_type => $token, @_);
	} else {
	    croak "bad direction";
	};
    };
    $self->{($direction eq 'input' ? '_in' : '_out')} = $jdb;
    $jdb->error and croak $self->{_prog} . ": cannot open $direction: " . $jdb->error . ".\n";
}

=head2 setup

    $filter->setup();

Do any setup that requires minimal IO
(for example, reading and parsing headers).

Called exactly once.

=cut

sub setup ($) {
    my $self = shift @_;
    die "Jdb:Filter: we expect the subprogram to handle setup.\n";
}

=head2 run

    $filter->run();

Execute the body, typically iterating over the input rows.

Called exactly once.

=cut

sub run ($) {
    my $self = shift @_;
    die "Jdb:Filter: we expect the subprogram to handle run.\n";
}

=head2 finish

    $filter->finish();

Write out any trailing comments and close output.

=cut

sub finish ($) {
    my $self = shift @_;

    if (defined($self->{_delay_comments})) {
	$self->{_delay_comments}->flush($self->{_out});
    };

    my $args = '';
    $args = " " . join(" ", @{$self->{_orig_argv}})
	if (defined($self->{_orig_argv}) && $#{$self->{_orig_argv}} != -1);
    $self->{_out}->write_comment("  | " . $self->{_prog} . $args)
	if ($self->{_logprog});
    $self->{_out}->close
	if ($self->{_close});
}

=head2 setup_run_finish

    $filter->setup_run_finish();

Shorthand for doing everything needed to run a command straightaway.

=cut

sub setup_run_finish ($) {
    my $self = shift @_;
    $self->setup();
    $self->run();
    $self->finish();
}

=head2 info

    $filter->info($INFOTYPE)

Return information about what the filter does.
Infotypes:

=over

=item input_type
Types of input accepted.
Raw types are: "jdbtext", "jdbobj", "jdb*", "text", or "none".

=item output_type
Type of output produced.
Same format as input_type.

=item input_count
Number of input streams (usually 1).

=item output_count
Number of input streams (usually 1).

=back

=cut

sub info ($) {
    my $self = shift @_;
    my $key = shift @_;
    return $self->{_info}{$key};
}


=head1 CLASS-SPECIFIC UTILITY ROUTINES

Filter has some class-specific utility routines in it.
(I.e., they know about $self.)

=head2 create_pass_comments_sub

    $filter->create_pass_comments_sub

Creates a code block suitable for passing to Jdb::IO::Readers -comment_handler
that passes comments through to $self->{_out}.

=cut

sub create_pass_comments_sub ($)
{
    my $self = shift @_;
#xxx    # one extra level of indirection to allow for delayed opening of _out
    return sub { $self->{_out}->write_raw(@_); };
}

=head2 create_delay_comments_sub

    $filter->create_delay_comments_sub

Creates a code block suitable for passing to Jdb::IO::Readers -comment_handler
that will buffer comments for automatic (from $self->final) after all other IO.

=cut

sub create_delay_comments_sub ($) {
    my $self = shift @_;
    my $dpc = $self->{_delay_comments} = new Jdb::Support::DelayPassComments(\$self->{_out});
    return sub { $dpc->enqueue(@_); };
}


=head2 create_compare_code

    $filter->create_compare_code($a_jdb, $b_jdb).

Write compare code based on sort-style options
stored in C<$self->{_sort_argv}>.
We assume C<$a> and C<$b> point to arefs.
$A_JDB and $B_JDB are the Jdb::IO object that defines the schemas
for $a and $b.

=cut

sub create_compare_code ($$) {
    my($self, $a_jdb, $b_jdb) = @_;

    my $compare_code = "sub {\n" . "\treturn\n";
    my ($reverse, $numeric) = (0, 0);
    my $arg;
    foreach $arg (@{$self->{_sort_argv}}) {
	if ($arg eq '-r') {
	    $reverse = 1;
	} elsif ($arg eq '-R') {
	    $reverse = 0;
	} elsif ($arg eq '-n') {
	    $numeric = 1;
	} elsif ($arg eq '-N') {
	    $numeric = 0;
        } elsif ($arg =~ /^-/) {
	    croak $self->{_prog} . ": internal error: unknown option $_ in sort key\n";
	} else {
	    my ($left, $right) = ($reverse ? (qw(b a)) : (qw(a b)));
	    my $left_coli = $a_jdb->col_to_i($arg);
	    my $right_coli = $b_jdb->col_to_i($arg);
	    if ($reverse) {
		my $tmp_coli = $left_coli;
		$left_coli = $right_coli;
		$right_coli = $tmp_coli;
	    };
	    croak $self->{_prog} . ": unknown column name $arg in sort key\n"
		if (!defined($left_coli) || !defined($right_coli));
	    $compare_code .= "\t" . '($' . $left . '->[' . $left_coli . '] ' .
    	    	    ($numeric ? "<=>" : "cmp") .
		    ' $' . $right . '->[' . $right_coli . ']) || ' .
		    ' # ' . $arg  .
		    ($reverse ? ", descending" : ", ascending") .
		    ($numeric ? " numeric" : " lexical") .
		    "\n";
	    # note that we don't currently handle NaN comparisons returning undef
	};
    };
    $compare_code .= "\t0; # match\n}\n";
    return $compare_code;
}

=head2 numeric_formatting

    $out = $self->numeric_formatting($x)

Display a floating point number $x using $self->{_format},
handling possible non-numeric "-" as a special case.

=cut

sub numeric_formatting {
    my ($self, $x) = @_;
    return $x if ($x eq '-');
    return sprintf($self->{_format}, $x);
}

=head2 setup_exactly_two_inputs

    $self->setup_exactly_two_inputs

Ensure that there are exactly two input streams.
Common to L<dbmerge> and L<dbjoin>.

=cut

sub setup_exactly_two_inputs {
    my($self) = @_;
    if ($#{$self->{_inputs}} == -1) {
	croak $self->{_prog} . ": too few input sources specified, use --input.\n";
    };
    if ($#{$self->{_inputs}} > 1) {
	croak $self->{_prog} . ": too input sources specified, dbmerge only hanldes two at once.\n";
    };
    if ($#{$self->{_inputs}} == 0) {
	# need to use stdin?
	my $token = new IO::Handle;
	$token->fdopen(fileno(STDIN), "r");
	unshift @{$self->{_inputs}}, $token;
    };
    croak if ($#{$self->{_inputs}} != 1);   # assert
}


=head1 NON-CLASS UTILITY ROUTINES

Filter also has some utility routines that are not part of the class structure.
They are not exported.

(none currently)

=cut

1;
