#!/usr/bin/perl -w

#
# dbjoin.pm
# Copyright (C) 1991-2007 by John Heidemann <johnh@isi.edu>
# $Id: dbjoin.pm 622 2008-01-26 05:46:57Z johnh $
#
# This program is distributed under terms of the GNU general
# public license, version 2.  See the file COPYING
# in $dblibdir for details.
#

package Jdb::Filter::dbjoin;

=head1 NAME

dbjoin - join two tables on common columns

=head1 SYNOPSIS

dbjoin [-Sid] table1 table2 [-nNrR] column [column...]

=head1 DESCRIPTION

Does a natural join on TABLE1 and TABLE2 
the specified columns.

By default, data will be sorted lexically,
but the usual sorting options can be mixed with the column
specification.

If data is already sorted, dbjoin will run more efficiently
with the C<-S> option.

Because two tables are required,
input is typically in files.
Standard input is accessible by the file "-".

=head1 OPTIONS

=over

=item C<-a> or C<--all>

include non-matches (each record which doesn't match at
all will appear once)

=item C<-S> or B<--pre-sorted>

assume (and verify) data is already sorted

=item C<-e E> or C<--empty E>
give value E as the value for empty (null) records

=item <-T TmpDir>

where to put tmp files.
Also uses environment variable TMPDIR, if -T is 
not specified.
Default is /tmp.

=back

Sort specification options (can be interspersed with column names):

=over

=item B<-r> or B<--descending>

sort in reverse order (high to low)

=item B<-R> or B<--ascending>

sort in normal order (low to high)

=item B<-n> or B<--numeric>

sort numerically

=item B<-N> or B<--lexical>

sort lexicographically

=back


=for comment
begin_standard_jdb_options

This module also supports the standard jdb options:

=over

=item B<-d>

Enable debugging output.

=item B<-i> or B<--input> InputSource

Read from InputSource, typically a file name, or C<-> for standard input,
or (if in Perl) a IO::Handle, Jdb::IO or Jdb::BoundedQueue objects.

=item B<-o> or B<--output> OutputDestination

Write to OutputDestination, typically a file name, or C<-> for standard output,
or (if in Perl) a IO::Handle, Jdb::IO or Jdb::BoundedQueue objects.

=item B<--autorun> or B<--noautorun>

By default, programs process automatically,
but Jdb::Filter objects in Perl do not run until you invoke
the run() method.
The C<--(no)autorun> option controls that behavior within Perl.

=item B<--help>

Show help.

=item B<--man>

Show full manual.

=back

=for comment
end_standard_jdb_options


=head1 SAMPLE USAGE

=head2 Input:

    #h sid cid
    1 10
    2 11
    1 12
    2 12

And in the file F<DATA/classes>:

    #h cid cname
    10 pascal
    11 numanal
    12 os

=head2 Command:

    cat DATA/reg.jdb | dbsort -n cid | dbjoin - DATA/classes -n cid

=head2 Output:

    #h      cid     sid     cname
    10      1       pascal
    11      2       numanal
    12      1       os
    12      2       os
    # - COMMENTS:
    #  | /home/johnh/BIN/DB/dbsort -n cid
    # DATA/classes COMMENTS:
    # joined comments:
    #  | /home/johnh/BIN/DB/dbjoin - DATA/classes cid

=head1 SEE ALSO

L<Jdb>.


=head1 CLASS FUNCTIONS

=cut

@ISA = qw(Jdb::Filter);
($VERSION) = ('$Revision: 622 $' =~ m/(\d+)/);

use strict;
use Pod::Usage;

use Jdb::Filter;
use Jdb::IO::Reader;
use Jdb::IO::Writer;


=head2 new

    $filter = new Jdb::Filter::dbjoin(@arguments);

Create a new dbjoin object, taking command-line arguments.

=cut

sub new ($@) {
    my $class = shift @_;
    my $self = $class->SUPER::new(@_);
    bless $self, $class;
    $self->set_defaults;
    $self->parse_options(@_);
    $self->SUPER::post_new();
    return $self;
}


=head2 set_defaults

    $filter->set_defaults();

Internal: set up defaults.

=cut

sub set_defaults ($) {
    my($self) = @_; 
    $self->SUPER::set_defaults();
    $self->{_include_non_matches} = 0;
    $self->{_pre_sorted} = 0;
}

=head2 parse_options

    $filter->parse_options(@ARGV);

Internal: parse command-line arguments.

=cut

sub parse_options ($@) {
    my $self = shift @_;

    my(@argv) = @_;
    $self->get_options(
	\@argv,
 	'help|?' => sub { pod2usage(1); },
	'man' => sub { pod2usage(-verbose => 2); },
	'a|all!' => \$self->{_include_non_matches},
	'autorun!' => \$self->{_autorun},
	'd|debug+' => \$self->{_debug},
	'e|empty=s' => \$self->{_empty},
	'i|input=s@' => sub { $self->parse_io_option('inputs', @_); },
	'log!' => \$self->{_logprog},
	'o|output=s' => sub { $self->parse_io_option('output', @_); },
	'S|pre-sorted+' => \$self->{_pre_sorted},
	'T|tmpdir|tempdir=s' => \$self->{_tmpdir},
	# sort key options:
	'n|numeric' => sub { $self->parse_sort_option(@_); },
	'N|lexical' => sub { $self->parse_sort_option(@_); },
	'r|descending' => sub { $self->parse_sort_option(@_); },
	'R|ascending' => sub { $self->parse_sort_option(@_); },
	'<>' => sub { $self->parse_sort_option('<>', @_); },
	) or pod2usage(2);
    croak $self->{_prog} . ": interal error, extra arguments.\n"
	if ($#argv != -1);
}

=head2 setup

    $filter->setup();

Internal: setup, parse headers.

=cut

sub setup ($) {
    my($self) = @_;

    croak $self->{_prog} . ": no sorting key specified.\n"
	if ($self->{_sort_argv} == -1);

    $self->setup_exactly_two_inputs;

    $self->finish_io_option('inputs', -comment_handler => $self->create_delay_comments_sub);
    $self->finish_io_option('output', -clone => $self->{_ins}[0]);

    $self->{_compare_code} = $self->create_compare_code(@{$self->{_ins}});;
    print "COMPARE CODE:\n\t" . $self->{_compare_code} . "\n" if ($self->{_debug});
}

=head2 run

    $filter->run();

Internal: run over each rows.

=cut
sub run ($) {
    my($self) = @_;

    # Override the usual package globals for $a, $b (!),
    # then eval compare_sub in this lexical context.
    # We're Having Fun Now.
    my($a, $b);
    my $compare_sub;
    eval '$compare_sub = ' . $self->{_compare_code};
    $@ && croak $self->{_prog} . ":  internal eval error in compare code: $@.\n";
    my $a_fastpath_sub = $self->{_ins}[0]->fastpath_sub();
    my $b_fastpath_sub = $self->{_ins}[1]->fastpath_sub();
    my $out_fastpath_sub = $self->{_out}->fastpath_sub();

    # prime the pump
    $a = &{$a_fastpath_sub}();
    $b = &{$b_fastpath_sub}();
    for (;;) {
	last if (!defined($a) || !defined($b));  # eof on one
	my $cmp = &{$compare_sub}();   # compare $a vs $b
	if ($cmp <= 0) {
	    &{$out_fastpath_sub}($a);
	    $a = &{$a_fastpath_sub}();
	} else {
	    &{$out_fastpath_sub}($b);
	    $b = &{$b_fastpath_sub}();
	};
    };
    # one should be done
    die if (defined($a) && defined($b));   # assert
    # drain the one that's still full
    while (defined($a)) {
	&{$out_fastpath_sub}($a);
	$a = &{$a_fastpath_sub}();
    };
    while (defined($b)) {
	&{$out_fastpath_sub}($b);
	$b = &{$b_fastpath_sub}();
    };
    foreach (0..1) {
	$self->{_ins}[$_]->close;
    };
}


die "xxx: old code below.\n";


if ($sorting_required) {
    # dbbindir
    open(B, "cat $bname | dbsort $join_columns |") || die("$prog: cannot run dbsort over ``$bname''.\n");
} else {
    open(B, "<$bname") || die("$prog: cannot open ``$bname''.\n");
};
my($bheader, @brawdata, @bcomments, @bdata, $b_fs_code, @bcoloptions, @bcolnames, @bsubcolnames, %bcolnametonum);
$bheader = <B>;
@brawdata = <B>;
@bcomments = grep(/^#/, @brawdata);
@bdata = grep(!/^#/, @brawdata);
chomp(@bdata);
undef @brawdata;
close (B);
&process_header($bheader);
$b_fs_code = $fs_code;
@bcoloptions = @coloptions;
@bcolnames = @colnames;
@bsubcolnames = @bcolnames;
%bcolnametonum = %colnametonum;
# reset
@colnames = ();
%colnametonum = ();

die "$prog: cannot handle input data with different field separators.\n"
    if ($a_fs_code ne $b_fs_code);
@coloptions = @acoloptions;   # NEEDSWORk

#
# figure the joined columns
#
$i = 0;
my %join_keys;
for $key (@ARGV) {
    next if ($key =~ /^-/);   # we deal with this later
    die("$prog: column ``$key'' is not in table ``$aname''.\n") if (!defined($acolnametonum{$key}));
    die("$prog: column ``$key'' is not in table ``$bname''.\n") if (!defined($bcolnametonum{$key}));
    $join_keys{$key} = $i;
    $colnames[$i] = $key;
    &col_mapping ($key, $i);
    $coltoa[$i] = $acolnametonum{$key};
    $coltob[$i] = $bcolnametonum{$key};
    $colina[$i] = 1;
    $colinb[$i] = 1;
    $i++;
}

#
# and the rest
#
foreach $key (@acolnames) {
    next if (defined($colnametonum{$key}));
    $colnames[$i] = $key;
    &col_mapping ($key, $i);
    $coltoa[$i] = $acolnametonum{$key};
    $colina[$i] = 1;
    $i++;
}
foreach $key (@bcolnames) {
    # detect duplicates that are not joined upon (error)
    # (this represents duplicate fieds in the two merged things).
    # Reject this because we don't want to silently prefer one to the other.
    if (defined($colnametonum{$key}) && !defined($join_keys{$key})) {
	die "$prog: column $key is in both of the joined files, but is not joined upon.\nAll non-joined columns must be unique.\nRename one of the source columns,\nor remove one of the duplicate input columns.\n";
    };
    next if (defined($colnametonum{$key}));
    $colnames[$i] = $key;
    &col_mapping ($key, $i);
    $coltob[$i] = $bcolnametonum{$key};
    $colinb[$i] = 1;
    $i++;
}

#
# Figure the comparison code.
#
sub join_row_col_ab_fn {
    my($row, $colname, $n) = @_;
    return '$' . $row . 'f[' . ($row eq 'a' ? $coltoa[$n] : $coltob[$n]) . ']';   # '
}
my($ab_compare_code) = &generate_compare_code('ab_custom_compare', '::join_row_col_ab_fn', @ARGV);
eval $ab_compare_code;
$@ && die "generate_compare_code: error ``$@ in'' eval of new code.\n\t$ab_compare_code\n";


sub join_row_col_aolda_fn {
    my($row, $colname, $n) = @_;
    return '$' . ($row eq 'a' ? 'a' : 'olda') . 'f[' . $coltoa[$n] . ']';   # '
}
my($aolda_compare_code) = &generate_compare_code('aolda_custom_compare', '::join_row_col_aolda_fn', @ARGV);
eval $aolda_compare_code;
$@ && die "generate_compare_code: error ``$@ in'' eval of new code.\n\t$aolda_compare_code\n";

sub join_row_col_boldb_fn {
    my($row, $colname, $n) = @_;
    return '$' . ($row eq 'a' ? 'b' : 'oldb') . 'f[' . $coltob[$n] . ']';   # '
}
my($boldb_compare_code) = &generate_compare_code('boldb_custom_compare', '::join_row_col_boldb_fn', @ARGV);
eval $boldb_compare_code;
$@ && die "generate_compare_code: error ``$@ in'' eval of new code.\n\t$boldb_compare_code\n";


&write_header();

#
# join the data (assumes data already sorted)
#
my($oldai, $oldbi, $ai, $bi, $firstmatchingbi, $lastmatchingbi, $inmatch);
$oldai = $oldbi = -1;   # keep track of the last entry to check for sortedness
$ai = $bi = 0;
$firstmatchingbi = $lastmatchingbi = -1;
    # $firstmatchingbi keeps track of the head of a run of bi's that match
    # $lastmatchingbi is different, it's the highestbi that's ever matched
$inmatch = 0;

sub save_ai {
    my($newval) = @_;
    @oldaf = @af;
    $oldai = $ai;
    $ai = $newval;
    @af = ($ai > $#adata ? () : split(/$fsre/, $adata[$ai]));
}

sub save_bi {
    my($newval) = @_;
    @oldbf = @bf;
    $oldbi = $bi;
    $bi = $newval;
    @bf = ($bi > $#bdata ? () : split(/$fsre/, $bdata[$bi]));
}

# loop through all a
&save_ai(0);
&save_bi(0);
while ($ai <= $#adata) {
    #
    # Get the two rows for comparision.
    #
    if ($bi > $#bdata) {
    	$result = -1;
    } else {
	# init sort checking
	if ($oldai == -1) {
	    @oldaf = @af;
	    $oldai = $ai;
	};
	if ($oldbi == -1) {
	    @oldbf = @bf;
	    $oldbi = $bi;
	};

	# check for sortedness
	print "# $ai($oldai) $bi($oldbi)\n" if ($debug);
	if ($oldai < $ai) {
#	    if (!defined($af[0])) {
#		warn "\$af[0] undef\n";
#	    };
#	    if (!defined($oldaf[0])) {
#	        warn "\$oldaf[0] undef\n";
#	    };
	    die("$prog: table ``$aname'' is out of order (between data lines $oldai and $ai).\n")
	        if (&aolda_custom_compare() < 0);
	};
	if ($oldbi < $bi) {
	    die("$prog: table ``$bname'' is out of order (between data lines $oldbi and $bi).\n")
	        if (&boldb_custom_compare() < 0);
	};

	$result = &ab_custom_compare();
	# Old way:
	# $result2 = &compare();
	# print "$result\t$result2\n";
    };
    if ($result != 0 && $inmatch) {
    	$inmatch = 0;
	# $ai has matched before, so don't need to check non-match inclusion
	&save_ai($ai+1);
	&save_bi($firstmatchingbi);
	next;
    };
    if ($result < 0) {
	if ($non_match_inclusion) {
	    &generate_a_only_data();
	    print "# a-only\n" if ($debug);
	    &print_joined_data();
	};
	&save_ai($ai+1);
    } elsif ($result > 0) {
	if ($non_match_inclusion && $bi > $lastmatchingbi) {
	    &generate_b_only_data();
	    print "# b-only\n" if ($debug);
	    &print_joined_data();
	};
	&save_bi($bi+1);
    } else {
    	if (!$inmatch) {
	    $inmatch = 1;
	    $firstmatchingbi = $bi;
	    $lastmatchingbi = $bi if ($bi > $lastmatchingbi);
	} else {
	    $lastmatchingbi = $bi if ($bi > $lastmatchingbi);
	};
	&generate_joined_data();
	&print_joined_data();
	&save_bi($bi+1);
    };
};
#
# When we're done, there could be b's left.
#
if ($non_match_inclusion) {
    while ($bi <= $#bdata) {
        if ($bi > $lastmatchingbi) {
            &generate_b_only_data();
	    print "# post-scan b-only\n" if ($debug);
	    &print_joined_data();
        };
        $bi++;
	&save_bi($bi);
    };
};

print "# $aname COMMENTS:\n" .
    join("", @acomments) .
    "# $bname COMMENTS:\n" .
    join("", @bcomments) .
    "# joined comments:\n";
print "#  | $prog ", join(" ", @orig_argv), "\n";
exit 0;

# for compiler warnings
my($x) = $outfs;
$x = $dbopts->optarg;

sub generate_joined_data {
    my($ci);
    @c = ();
    foreach $ci (0..$#colnames) {
	if ($colina[$ci]) {
	    push (@c, $af[$coltoa[$ci]]);
	} else {
	    push (@c, $bf[$coltob[$ci]]);
	};
    };
}

sub generate_a_only_data {
    my($ci);
    @c = ();
    foreach $ci (0..$#colnames) {
	if ($colina[$ci]) {
	    push (@c, $af[$coltoa[$ci]]);
	} else {
	    die ("$prog: need empty value, specify with option -e Value.\n") if (!defined($null_value) && $non_match_inclusion);
	    push (@c, $null_value);
	};
    };
}

sub generate_b_only_data {
    my($ci);
    @c = ();
    foreach $ci (0..$#colnames) {
	if ($colinb[$ci]) {
	    push (@c, $bf[$coltob[$ci]]);
	} else {
	    die ("$prog: need empty value, specify with -e Value.\n") if (!defined($null_value) && $non_match_inclusion);
	    push (@c, $null_value);
	};
    };
}

sub print_joined_data {
    print join($outfs, @c) . "\n";
}

sub compare {
    my($result);
    my($i) = 0;
    foreach (@ARGV) {
	next if (/^-/);
    	$result = $af[$coltoa[$i]] cmp $bf[$coltob[$i]];
	return $result if ($result != 0);
	$i++;
    };
    return 0;
}



=head2 finish

    $filter->finish();

Internal: write trailer.

=cut
sub finish ($) {
    my($self) = @_;

    $self->SUPER::finish();
}

=head1 AUTHOR and COPYRIGHT

Copyright (C) 1991-2007 by John Heidemann <johnh@isi.edu>

This program is distributed under terms of the GNU general
public license, version 2.  See the file COPYING
with the distribution for details.

=cut

1;
