#!/usr/bin/perl -w

#
# dbmerge.pm
# Copyright (C) 1991-2007 by John Heidemann <johnh@isi.edu>
# $Id: dbmerge.pm 622 2008-01-26 05:46:57Z johnh $
#
# This program is distributed under terms of the GNU general
# public license, version 2.  See the file COPYING
# in $dblibdir for details.
#

package Jdb::Filter::dbmerge;

=head1 NAME

dbmerge - merge two inputs in sorted order based on the the specified columns

=head1 SYNOPSIS

    dbmerge --input A.jdb --input B.jdb [-T TemporaryDirectory] [-nNrR] column [column...]

or
    cat A.jdb | dbmerge --input B.jdb [-T TemporaryDirectory] [-nNrR] column [column...]

=head1 DESCRIPTION

Merge exactly two sorted input files, producing one sorted result.
Inputs can both be specified with C<--input>, or one can come
from standard input and the other from C<--input>.

Dbmerge consumes a fixed amount of memory regardless of input size.

=head1 OPTIONS

General option:

=over

=item <-T TmpDir>

where to put tmp files.
Also uses environment variable TMPDIR, if -T is 
not specified.
Default is /tmp.

=back

Sort specification options (can be interspersed with column names):

=over

=item B<-r> or B<--descending>

sort in reverse order (high to low)

=item B<-R> or B<--ascending>

sort in normal order (low to high)

=item B<-n> or B<--numeric>

sort numerically

=item B<-N> or B<--lexical>

sort lexicographically

=back

=for comment
begin_standard_jdb_options

This module also supports the standard jdb options:

=over

=item B<-d>

Enable debugging output.

=item B<-i> or B<--input> InputSource

Read from InputSource, typically a file name, or C<-> for standard input,
or (if in Perl) a IO::Handle, Jdb::IO or Jdb::BoundedQueue objects.

=item B<-o> or B<--output> OutputDestination

Write to OutputDestination, typically a file name, or C<-> for standard output,
or (if in Perl) a IO::Handle, Jdb::IO or Jdb::BoundedQueue objects.

=item B<--autorun> or B<--noautorun>

By default, programs process automatically,
but Jdb::Filter objects in Perl do not run until you invoke
the run() method.
The C<--(no)autorun> option controls that behavior within Perl.

=item B<--help>

Show help.

=item B<--man>

Show full manual.

=back

=for comment
end_standard_jdb_options


=head1 SAMPLE USAGE

=head2 Input:

File F<a.jdb>:

    #h cid cname
    11 numanal
    10 pascal

File F<b.jdb>:

    #h cid cname
    12 os
    13 statistics

=head2 Command:

    dbmerge --input a.jdb --input b.jdb cname

or

    cat a.jdb | dbmerge --input b.jdb cname

=head2 Output:

    #h      cid     cname
    11 numanal
    12 os
    10 pascal
    13 statistics
    #  | dbmerge --input a.jdb --input b.jdb cname

=head1 SEE ALSO

L<dbsort(1)>,
L<Jdb(3)>

=head1 CLASS FUNCTIONS

=cut


@ISA = qw(Jdb::Filter);
($VERSION) = ('$Revision: 622 $' =~ m/(\d+)/);

use strict;
use Carp qw(croak);
use Pod::Usage;

use Jdb::Filter;
use Jdb::IO::Reader;
use Jdb::IO::Writer;

=head2 new

    $filter = new Jdb::Filter::dbmerge(@arguments);

Create a new object, taking command-line arugments.

=cut

sub new ($@) {
    my $class = shift @_;
    my $self = $class->SUPER::new(@_);
    bless $self, $class;
    $self->set_defaults;
    $self->parse_options(@_);
    $self->SUPER::post_new();
    return $self;
}


=head2 set_defaults

    $filter->set_defaults();

Internal: set up defaults.

=cut

sub set_defaults ($) {
    my $self = shift @_;
    $self->SUPER::set_defaults();
    $self->{_info}{input_count} = 2;
    $self->{_sort_argv} = [];
    $self->{_tmpdir} = defined($ENV{'TMPDIR'}) ? $ENV{'TMPDIR'} : "/tmp";
}

=head2 parse_options

    $filter->parse_options(@ARGV);

Internal: parse command-line arguments.

=cut

sub parse_options ($@) {
    my $self = shift @_;

    my(@argv) = @_;
    $self->get_options(
	\@argv,
 	'help|?' => sub { pod2usage(1); },
	'man' => sub { pod2usage(-verbose => 2); },
	'autorun!' => \$self->{_autorun},
	'close!' => \$self->{_close},
	'd|debug+' => \$self->{_debug},
	'i|input=s@' => sub { $self->parse_io_option('inputs', @_); },
	'log!' => \$self->{_logprog},
	'o|output=s' => sub { $self->parse_io_option('output', @_); },
	'T|tmpdir|tempdir=s' => \$self->{_tmpdir},
	# sort key options:
	'n|numeric' => sub { $self->parse_sort_option(@_); },
	'N|lexical' => sub { $self->parse_sort_option(@_); },
	'r|descending' => sub { $self->parse_sort_option(@_); },
	'R|ascending' => sub { $self->parse_sort_option(@_); },
	'<>' => sub { $self->parse_sort_option('<>', @_); },
	) or pod2usage(2);
}


=head2 setup

    $filter->setup();

Internal: setup, parse headers.

=cut

sub setup ($) {
    my($self) = @_;

    croak $self->{_prog} . ": no sorting key specified.\n"
	if ($self->{_sort_argv} == -1);

    #
    # setup final IO
    #
    $self->setup_exactly_two_inputs;
    $self->finish_io_option('inputs', -comment_handler => $self->create_pass_comments_sub);
    $self->finish_io_option('output', -clone => $self->{_ins}[0]);

    croak $self->{_prog} . ": intput streams don't have identical schemas\n"
	if ($self->{_ins}[0]->compare($self->{_ins}[1]) ne 'identical');

    $self->{_compare_code} = $self->create_compare_code(@{$self->{_ins}});;
    print "COMPARE CODE:\n\t" . $self->{_compare_code} . "\n" if ($self->{_debug});
}


=head2 run

    $filter->run();

Internal: run over each rows.

=cut
sub run ($) {
    my($self) = @_;

    # Override the usual package globals for $a, $b (!),
    # then eval compare_sub in this lexical context.
    # We're Having Fun Now.
    my($a, $b);
    my $compare_sub;
    eval '$compare_sub = ' . $self->{_compare_code};
    $@ && croak $self->{_prog} . ":  internal eval error in compare code: $@.\n";
    my $a_fastpath_sub = $self->{_ins}[0]->fastpath_sub();
    my $b_fastpath_sub = $self->{_ins}[1]->fastpath_sub();
    my $out_fastpath_sub = $self->{_out}->fastpath_sub();

    # prime the pump
    $a = &{$a_fastpath_sub}();
    $b = &{$b_fastpath_sub}();
    for (;;) {
	last if (!defined($a) || !defined($b));  # eof on one
	my $cmp = &{$compare_sub}();   # compare $a vs $b
	if ($cmp <= 0) {
	    &{$out_fastpath_sub}($a);
	    $a = &{$a_fastpath_sub}();
	} else {
	    &{$out_fastpath_sub}($b);
	    $b = &{$b_fastpath_sub}();
	};
    };
    # one should be done
    die if (defined($a) && defined($b));   # assert
    # drain the one that's still full
    while (defined($a)) {
	&{$out_fastpath_sub}($a);
	$a = &{$a_fastpath_sub}();
    };
    while (defined($b)) {
	&{$out_fastpath_sub}($b);
	$b = &{$b_fastpath_sub}();
    };
    foreach (0..1) {
	$self->{_ins}[$_]->close;
    };
};
    

=head1 AUTHOR and COPYRIGHT

Copyright (C) 1991-2007 by John Heidemann <johnh@isi.edu>

This program is distributed under terms of the GNU general
public license, version 2.  See the file COPYING
with the distribution for details.

=cut

1;
