#!/usr/bin/perl -w

#
# dbpipeline.pm
# Copyright (C) 2007 by John Heidemann <johnh@isi.edu>
# $Id: dbpipeline.pm 620 2007-12-23 06:08:15Z johnh $
#
# This program is distributed under terms of the GNU general
# public license, version 2.  See the file COPYING
# in $dblibdir for details.
#

package Jdb::Filter::dbpipeline;

=head1 NAME

dbpipeline - allow db commands to be assembled as pipelines in Perl

=head1 SYNOPSIS

    use Jdb::Filter::dbpipeline qw(:all);
    dbpipeline(
        dbrow(qw(name test1)),
        dbroweval('_test1 += 5;')
    );

=head1 DESCRIPTION

This module makes it easy to create threaded pipelines in Perl.

By default (as with all Jdb modules), input is from STDIN and output
to STDOUT.  Two helper functions, fromfile and tofile can grab
data from files.

Dbpipeline differs in several ways from all other Jdb::Filter modules:
it does not have a corresponding Unix command (it is used only from
within Perl).
It does not log its presence to the output stream (this is arguably a bug,
but it doesn't actually do anything).


=head1 OPTIONS

Unlike most Jdb modules, dbpipeline defaults to C<--autorun>.

=for comment
begin_standard_jdb_options

This module also supports the standard jdb options:

=over

=item B<-d>

Enable debugging output.

=item B<-i> or B<--input> InputSource

Read from InputSource, typically a file name, or C<-> for standard input,
or (if in Perl) a IO::Handle, Jdb::IO or Jdb::BoundedQueue objects.

=item B<-o> or B<--output> OutputDestination

Write to OutputDestination, typically a file name, or C<-> for standard output,
or (if in Perl) a IO::Handle, Jdb::IO or Jdb::BoundedQueue objects.

=item B<--autorun> or B<--noautorun>

By default, programs process automatically,
but Jdb::Filter objects in Perl do not run until you invoke
the run() method.
The C<--(no)autorun> option controls that behavior within Perl.

=item B<--help>

Show help.

=item B<--man>

Show full manual.

=back

=for comment
end_standard_jdb_options


=head1 SEE ALSO

L<Jdb(3)>


=head1 CLASS FUNCTIONS

=cut

@ISA = qw(Jdb::Filter);
($VERSION) = ('$Revision: 620 $' =~ m/(\d+)/);

use Exporter 'import';
@EXPORT = ();
@EXPORT_OK = qw(
    dbpipeline_sink
    dbpipeline_open2
    dbpipeline_close2_hash
    );
# update them here, in toplevel Makefile.PL, and below in the documentation.
our @modularized_programs = qw(
	dbcol
	dbcolcreate
	dbcolrename
	dbcolstats
	dbcolstatscores
	dbmerge
	dbpipeline
	dbrow
	dbrowcount
	dbroweval
	dbsort
	);
%EXPORT_TAGS = (all => [@EXPORT_OK, @modularized_programs]);
Exporter::export_ok_tags('all');
my %autoloadable = map { $_ => 1 } @modularized_programs;

use strict;
use threads;
use Carp;
use Pod::Usage;

use Jdb::BoundedQueue;
use Jdb::Filter;
use Jdb::IO::Reader;
use Jdb::IO::Writer;

#
# First off, create all the bindings we promise in EXPORT_TAGS.
# Automated via AUTOLOAD for extra coolness.
#
our $AUTOLOAD;
sub AUTOLOAD {
    my $sub = $AUTOLOAD;
    (my $localsub = $sub) =~ s/.*:://;
    die "dbpipeline: AUTOLOAD on non-autoloadable sub $sub\n"
	if (!defined($autoloadable{$localsub}));
    eval "sub $localsub { use Jdb::Filter::$localsub; return new Jdb::Filter::$localsub(" . '@_' . "); };\n";
    $@ and die "dbpipeline: error creating stubs: $@\n";
    goto &$sub;
}

sub DESTROY {
    # just suppress autoloading warnings
}

=head2 dbpipeline

    dbpipeline(@modules);

This shorthand-routine creates a dbpipeline object
and then I<immediately runs it>.

Thus perl code becomes nearly as terse as shell code:

    dbpipeline(
	dbcol(qw(name test1)),
	dbroweval('_test1 += 5;'),
    );

The following commands currently have shorthand aliases:

=over

=item L<dbcol(1)>

=item L<dbcolcreate(1)>

=item L<dbcolrename(1)>

=item L<dbcolstats(1)>

=item L<dbcolstatscores(1)>

=item L<dbmerge(1)>

=item L<dbrow(1)>

=item L<dbrowcount(1)>

=item L<dbroweval(1)>

=item L<dbsort(1)>

=back

=cut

=head2 dbpipeline_sink

    my($jdb_writer, $thread) = dbpipeline_sink($writer_arguments_aref, @modules);

Set up a pipeline of @MODULES that is a data "sink", where the output
is given by a C<--output> argument, or goes to standard output (by default).
The caller generates input into the pipeline 
by writing to a newly created $JDB_WRITER,
whose configuration is specified by the manditory first
arugment $WRITER_ARGUMENTS_AREF.
(These arguments should include the schema.)
Returns this writer, and a $THREAD that must be joined to guarantee output
has completed.

=cut

sub dbpipeline_sink ($@) {
    my($writer_aref) = shift @_;
    
    my $queue = new Jdb::BoundedQueue;  # buffer from caller into pipeline
    my $writer = new Jdb::IO::Writer(-queue => $queue, @$writer_aref);

    my $pipeline = new Jdb::Filter::dbpipeline(
		    '--noautorun',
		    '--input' => $queue,
		    @_);
    my $pipeline_thread = threads->new(
	sub {
	    $pipeline->setup_run_finish;
	} );
    return ($writer, $pipeline_thread);
}

=head2 dbpipeline_open2

    my($jdb_read_queue, $jdb_writer, $thread) = dbpipeline_open2($writer_arguments_aref, @modules);

Set up a pipeline of @MODULES that is a data sink and source (both!).
The caller generates input into the pipeline 
by writing to a newly created $JDB_WRITER,
whose configuration is specified by the manditory 
arugment $WRITER_ARGUMENTS_AREF.
(These arguments should include the schema.)
The output of the pipeline comes out to the newly
created $JDB_READ_QUEUE.
Returns this read queue and writer,
and a $THREAD that must be joined to guarantee output
has completed.

(Unfortunately the interface is assymetric with a read I<queue>
but a write C<Jdb::IO> object, because C<Jdb::IO::Reader> blocks on
input of the header.

Like L<IPC::Open2>, with all of its pros and cons like potential deadlock.

=cut

sub dbpipeline_open2 ($@) {
    my($writer_aref) = shift @_;
    
    my $into_pipeline_queue = new Jdb::BoundedQueue;
    my $from_pipeline_queue = new Jdb::BoundedQueue;
    my $writer = new Jdb::IO::Writer(-queue => $into_pipeline_queue, @$writer_aref);
    # can't make a reader here since it will block on the header
#    my $reader = new Jdb::IO::Reader(-queue => $from_pipeline_queue, @$reader_aref);

    my $pipeline = new Jdb::Filter::dbpipeline(
		    '--noautorun',
		    '--input' => $into_pipeline_queue,
		    '--output' => $from_pipeline_queue,
		    @_);
    my $pipeline_thread = threads->new(
	sub {
	    $pipeline->setup_run_finish;
	} );
    return ($from_pipeline_queue, $writer, $pipeline_thread);
}

=head2 dbpipeline_close2_hash

    my($href) = dbpipeline_close2_hash($jdb_read_queue, $jdb_writer, $thread);

Reads and returns one line of output from $JDB_READER,
after closing $JDB_WRITER and joining the $THREAD.

Useful, for example, to get L<dbcolstats> output cleanly.

=cut

sub dbpipeline_close2_hash ($$$) {
    my($read_queue, $writer, $thr) = @_;
    $writer->close;
    if (defined($thr)) {
	$thr->join;
    };
    my %out_hash;
    my $reader = new Jdb::IO::Reader(-queue => $read_queue);
    $reader->error and croak "dbpipeline_close2_hash: couldn't setup reader.\n";
    $reader->read_row_to_href(\%out_hash) or croak "dbpipeline_close2_hash: no output from pipeline.\n";
    # check for eof
    my $o;
    while ($o = $reader->read_rowobj) {
	next if (!ref($o));  # comment
	# data is bad
	$o and croak "dbpipeline_close2_hash: multiple lines of output.\n";
    };
    return \%out_hash;
}



=head2 new

    $filter = new Jdb::Filter::dbpipeline(@arguments);

=cut

sub new {
    my $class = shift @_;
    my $self = $class->SUPER::new(@_);
    bless $self, $class;
    $self->set_defaults;
    $self->parse_options(@_);
    $self->SUPER::post_new();
    return $self;
}


=head2 set_defaults

    $filter->set_defaults();

Internal: set up defaults.

=cut

sub set_defaults ($) {
    my($self) = @_;
    $self->{_modules} = [];
    $self->SUPER::set_defaults();
    $self->{_autorun} = 1;  # override superclass default
}

=head2 parse_options

    $filter->parse_options(@ARGV);

Internal: parse options

=cut

sub parse_options ($@) {
    my $self = shift @_;

    my @argv = @_;
    $self->get_options(
	\@argv,
 	'help|?' => sub { pod2usage(1); },
	'man' => sub { pod2usage(-verbose => 2); },
	'autorun!' => \$self->{_autorun},
	'd|debug+' => \$self->{_debug},
	'i|input=s' => sub { $self->parse_io_option('input', @_); },
	'log!' => \$self->{_logprog},
	'o|output=s' => sub { $self->parse_io_option('output', @_); },
	) or pod2usage(2);
    push(@{$self->{_modules}}, @argv);
}

=head2 setup

    $filter->setup();

Internal: setup, parse headers.

=cut

sub setup ($) {
    my($self) = @_;

    my $last_module_i = $#{$self->{_modules}};
    die $self->{_prog} . ": no modules in pipeline.\n"
        if ($last_module_i < 0);

    my @queues;
    #
    # Make sure module inputs are sensible
    # and link their IO together, if so.
    #
    my $prev_mod = undef;
    my $i = 1;
    my $mod;
    foreach $mod (@{$self->{_modules}}) {
	die $self->{_prog} . ": module $i isn't type Jdb::Filter.\n"
	    if (ref($mod) !~ /^Jdb::Filter/);
	if (defined($prev_mod)) {
	    die $self->{_prog} . ": incompatible module input and output between modules $i and " . $i+1 . ".\n"
		if ($prev_mod->info('output_type') ne $mod->info('input_type'));
	    # xxx: above is a bit too strict, since jdbtext should match jdb*

	    my $queue = new Jdb::BoundedQueue;
	    $prev_mod->parse_options("--output" => $queue);
	    $mod->parse_options("--input" => $queue);
	    push(@queues, $queue);   # just for yucks
	};

	$prev_mod = $mod;
	$i++;
    };
    # Pass our input and output to the first and last modules:
    $self->{_modules}[0]->parse_options('--input' => $self->{_input})
	if (defined($self->{_input}));
    $self->{_modules}[$last_module_i]->parse_options('--output' => $self->{_output})
	if (defined($self->{_output}));

    # setup children
    foreach $mod (@{$self->{_modules}}) {
	$mod->setup();
    };

}

=head2 run

    $filter->run();

Internal: run over all IO

=cut
sub run ($) {
    my($self) = @_;
    my @threads = ();
    foreach my $mod (@{$self->{_modules}}) {
	# Note that we have to execute BOTH run and finish,
	# otherwise the first module never sends EOF down the queue.
	push(@threads, threads->new( sub { $mod->run; $mod->finish; } ));
    };
    # wait for termination of everyone
    foreach (@threads) {
	$_->join;
    };
}

=head2 finish

    $filter->finish();

Internal: we would write a trailer, but we don't because
we depend on the last command in the pipeline to do that.
We don't actually have a valid output stream.

=cut
sub finish ($) {
    my($self) = @_;
    # do nothing, since we have no output path.
    # $self->SUPER::finish();
}

=head1 AUTHOR and COPYRIGHT

Copyright (C) 1991-2007 by John Heidemann <johnh@isi.edu>

This program is distributed under terms of the GNU general
public license, version 2.  See the file COPYING
with the distribution for details.

=cut

1;

