#!/usr/bin/perl -w

#
# dbsort.pm
# Copyright (C) 1991-2007 by John Heidemann <johnh@isi.edu>
# $Id: dbsort.pm 622 2008-01-26 05:46:57Z johnh $
#
# This program is distributed under terms of the GNU general
# public license, version 2.  See the file COPYING
# in $dblibdir for details.
#

package Jdb::Filter::dbsort;

=head1 NAME

dbsort - sort rows based on the the specified columns

=head1 SYNOPSIS

    dbsort [-M MemLimit] [-T TemporaryDirectory] [-nNrR] column [column...]

=head1 DESCRIPTION

Sort all input rows as specified by the numeric or lexical columns.

Dbsort consumes a fixed amount of memory regardless of input size.
(It reverts to temporary files on disk if necessary, based on the -M
and -T options.)

The sort should be stable, but this has not yet been verified.

=head1 OPTIONS

General option:

=over

=item <-M MaxMemBytes>

Specify an approximate limit on memory usage (in bytes).
Larger values allow faster sorting because more operations
happen in-memory, provided you have enough memory.

=item <-T TmpDir>

where to put tmp files.
Also uses environment variable TMPDIR, if -T is 
not specified.
Default is /tmp.

=back

Sort specification options (can be interspersed with column names):

=over

=item B<-r> or B<--descending>

sort in reverse order (high to low)

=item B<-R> or B<--ascending>

sort in normal order (low to high)

=item B<-n> or B<--numeric>

sort numerically

=item B<-N> or B<--lexical>

sort lexicographically

=back

=for comment
begin_standard_jdb_options

This module also supports the standard jdb options:

=over

=item B<-d>

Enable debugging output.

=item B<-i> or B<--input> InputSource

Read from InputSource, typically a file name, or C<-> for standard input,
or (if in Perl) a IO::Handle, Jdb::IO or Jdb::BoundedQueue objects.

=item B<-o> or B<--output> OutputDestination

Write to OutputDestination, typically a file name, or C<-> for standard output,
or (if in Perl) a IO::Handle, Jdb::IO or Jdb::BoundedQueue objects.

=item B<--autorun> or B<--noautorun>

By default, programs process automatically,
but Jdb::Filter objects in Perl do not run until you invoke
the run() method.
The C<--(no)autorun> option controls that behavior within Perl.

=item B<--help>

Show help.

=item B<--man>

Show full manual.

=back

=for comment
end_standard_jdb_options


=head1 SAMPLE USAGE

=head2 Input:

    #h cid cname
    10 pascal
    11 numanal
    12 os

=head2 Command:

    cat data.jdb | dbsort cname

=head2 Output:

    #h      cid     cname
    11 numanal
    12 os
    10 pascal
    #  | dbsort cname

=head1 SEE ALSO

L<dbmerge(1)>,
L<dbmapreduce(1)>,
L<Jdb(3)>

=head1 CLASS FUNCTIONS

=cut


@ISA = qw(Jdb::Filter);
($VERSION) = ('$Revision: 622 $' =~ m/(\d+)/);

use strict;
use Pod::Usage;
use Carp;

use Jdb::IO::Reader;
use Jdb::IO::Writer;
use Jdb::Filter;
use Jdb::Filter::dbmerge;
use Jdb::Support::NamedTmpfile;

my($PERL_MEM_OVERHEAD) = 50;  # approx. bytes of overhead for each record in mem
my($PERL_MEM_SCALING) = 2;    # divided user requested mem by this factor to account to perl memory usage (huge approximatation)


=head2 new

    $filter = new Jdb::Filter::dbsort(@arguments);

Create a new object, taking command-line arugments.

=cut

sub new ($@) {
    my $class = shift @_;
    my $self = $class->SUPER::new(@_);
    bless $self, $class;
    $self->set_defaults;
    $self->parse_options(@_);
    $self->SUPER::post_new();
    return $self;
}


=head2 set_defaults

    $filter->set_defaults();

Internal: set up defaults.

=cut

sub set_defaults ($) {
    my $self = shift @_;
    $self->SUPER::set_defaults();
    $self->{_max_memory} = 1024*1024*64;
    $self->{_mem_debug} = undef;
    $self->{_sort_argv} = [];
    $self->{_tmpdir} = defined($ENV{'TMPDIR'}) ? $ENV{'TMPDIR'} : "/tmp";
}

=head2 parse_options

    $filter->parse_options(@ARGV);

Internal: parse command-line arguments.

=cut

sub parse_options ($@) {
    my $self = shift @_;

    my(@argv) = @_;
    $self->get_options(
	\@argv,
 	'help|?' => sub { pod2usage(1); },
	'man' => sub { pod2usage(-verbose => 2); },
	'autorun!' => \$self->{_autorun},
	'd|debug+' => \$self->{_debug},
	'i|input=s' => sub { $self->parse_io_option('input', @_); },
	'log!' => \$self->{_logprog},
	'M|maxmemory=i' => \$self->{_max_memory},
	'o|output=s' => sub { $self->parse_io_option('output', @_); },
	'T|tmpdir|tempdir=s' => \$self->{_tmpdir},
	# sort key options:
	'n|numeric' => sub { $self->parse_sort_option(@_); },
	'N|lexical' => sub { $self->parse_sort_option(@_); },
	'r|descending' => sub { $self->parse_sort_option(@_); },
	'R|ascending' => sub { $self->parse_sort_option(@_); },
	'<>' => sub { $self->parse_sort_option('<>', @_); },
	) or pod2usage(2);
    croak $self->{_prog} . ": interal error, extra arguments.\n"
	if ($#argv != -1);
}


=head2 setup

    $filter->setup();

Internal: setup, parse headers.

=cut

sub setup ($) {
    my($self) = @_;

    croak $self->{_prog} . ": no sorting key specified.\n"
	if ($self->{_sort_argv} == -1);

    #
    # setup final IO
    #
    $self->finish_io_option('input', -comment_handler => $self->create_delay_comments_sub);

    $self->{_compare_code} = $self->create_compare_code($self->{_in}, $self->{_in});
    print "COMPARE CODE:\n\t" . $self->{_compare_code} . "\n" if ($self->{_debug});
    my $compare_sub;
    eval '$self->{_compare_sub} = $compare_sub = ' . $self->{_compare_code} . ';';
    $@ && croak $self->{_prog} . ":  internal eval error in compare code: $@.\n";
#    $@ && croak "dbsort.pm:  internal eval error in compare code: $@.\n";
}

=head2 segment_start

    $self->segment_start(\@rows);

Sorting happens internally,
to handle large things in pieces if necessary.

call C<$self->segment_start> to init things and to restart after an overflow
C<$self->segment_overflow> to close one segment and start the next,
and C<$self->segment_merge> to put them back together again.

Note that we don't invoke the merge code unless the data
exceeds some threshold size, so small sorts happen completely
in memory.

Once we give up on memory, all the merging happens by making
passes over the disk files.

=cut

sub segment_start ($\@) {
    my ($self, $rows_ref) =  @_;

    $#{$rows_ref} = -1;  # truncate array
}

=head2 segment_next_output

    $out = $self->segment_next_output($input_finished)

Internal: return a Jdb::IO::Writer as $OUT
that either points to our output or a temporary file, 
depending on how things are going.

=cut

sub segment_next_output {
    my ($self, $input_finished) = @_;
    my $final_output = ($#{$self->{_files_to_merge}} == -1 && $input_finished);
    my $out;
    if ($final_output) {
	$self->finish_io_option('output', -clone => $self->{_in});
        $out = $self->{_out};
	print "# final output\n" if ($self->{_debug});
    } else {
	# dump to a file for merging
	my $tmpfile = Jdb::Support::NamedTmpfile::alloc($self->{_tmpdir});
	$out = $tmpfile;   # just return the name
	push(@{$self->{_files_to_merge}}, $tmpfile);
	print "# intermediate file: $tmpfile\n" if ($self->{_debug});
    };
    return ($out, $final_output);
}

=head2 segment_overflow

    $self->segment_overflow(\@rows, $input_finished)

Called to sort @ROWS, writing them to the appropriate place.
$INPUT_FINISHED is set if all input has been read.

=cut

#sub so1 {
#    my ($self, $rows_ref) = @_;
#    my(@sorted_rows) = sort { $a->[0] <=> $b->[0] } @{$rows_ref};
#    return @sorted_rows;
#}

sub segment_overflow ($\@$) {
    my($self, $rows_ref, $input_finished) = @_;

    my $compare_sub = $self->{_compare_sub};
    my(@sorted_rows) = sort $compare_sub @{$rows_ref};
#    my(@sorted_rows) = $self->so1($rows_ref);

    my ($out_fn, $final_output) = $self->segment_next_output($input_finished, 'Jdb:IO');
    my $out;
    if (ref($out_fn) =~ /^Jdb::IO::Writer/) {
	$out = $out_fn;   # a bit hacky, but whatever
    } else {
	$out = new Jdb::IO::Writer(-file => $out_fn, -clone => $self->{_in});
    };

    my $write_fastpath_sub = $out->fastpath_sub;
    foreach (@sorted_rows) {
	&$write_fastpath_sub($_);
    };

    if (!$final_output) {
	$out->close;
    };

    $self->segment_start($rows_ref);
}

=head2 segment_merge

    $self->segment_merge();

Merge queued files, if any.

We process the work queue in
a file-system-cache-friendly order (based on ideas from
"Information and Control in Gray-box Systems" by
the Arpaci-Dusseau's at SOSP 2001.

Idea:  each "pass" through the queue, reverse the processing
order so that the most recent data (that's hopefully
in the file system cache in memory) is handled first.

This algorithm isn't perfect (sometimes if there's an odd number
of files in the queue you reach way back in time, but most of 
the time it's quite good).

Also, in an ideal world $max_mem actually would be some sizable
percentage of memory, and so this whole optimization would
be useless because there would be no spare memory for the file system
cache.  But for saftey reasons (because we don't know how much
RAM there is, and there is multiprocessing, etc.), $max_mem
is almost always hugely conservative.  
In Nov. 2001 it was 10MB, but most workstations have >= 512MB memory;
in Nov. 2007 it is 64MB, but most computers have 2GB of momory.

=cut

sub segment_merge ($) {
    my($self) = @_;
    return if ($#{$self->{_files_to_merge}} == -1);

    my $files_to_merge_ref = $self->{_files_to_merge};

    # keep track of how often to reverse
    my($files_before_reversing_queue) = 0;
    # Merge the files in a binary tree.
    while ($#{$files_to_merge_ref} >= 0) {
	# Each "pass", reverse the queue to reuse stuff in memory.
	if ($files_before_reversing_queue <= 0) {
	    @{$files_to_merge_ref} = reverse @{$files_to_merge_ref};
	    $files_before_reversing_queue = $#{$files_to_merge_ref} + 1;
	    print "# reversed queue, $files_before_reversing_queue before next reverse.\n" if ($self->{_debug});
	};
	$files_before_reversing_queue -= 2;
	# Pick off the two next files for merging.
	my($a_fn, $b_fn);
	die $self->{_prog} . ": segment_merge, odd number of segments.\n" if ($#{$files_to_merge_ref} == 0);   # assert
	$a_fn = shift @{$files_to_merge_ref};
	$b_fn = shift @{$files_to_merge_ref};
	if (($a_fn cmp $b_fn) > 0) {
	    # Take files in order with the goal of providing a stable sort.
	    my $tmp_fn = $a_fn;
	    $a_fn = $b_fn;
	    $b_fn = $tmp_fn;
	};
	my($out_fn, $final_output) = $self->segment_next_output(1);
	print "# merging $a_fn and $b_fn\n" if ($self->{_debug});

	my @merge_options = qw(--autorun --nolog);
	push(@merge_options, '--noclose') if ($final_output);
	new Jdb::Filter::dbmerge(@merge_options,
				'--input' => $a_fn, '--input' => $b_fn,
				'--output' => $out_fn,
				@{$self->{_sort_argv}});

	Jdb::Support::NamedTmpfile::cleanup_one($a_fn);
	Jdb::Support::NamedTmpfile::cleanup_one($b_fn);
    };
}


=head2 run

    $filter->run();

Internal: run over each rows.

=cut
sub run ($) {
    my($self) = @_;

    #
    # read in and set up the data
    #
    $self->{_files_to_merge} = [];
    my $read_fastpath_sub = $self->{_in}->fastpath_sub();

    my $fref;    # the current row
    my @rows;      # an array of frefs for each row, $i long
    my $memory_used = 0;
    my $scaled_max_memory =  int($self->{_max_memory} / (1.0 * $PERL_MEM_SCALING));
    my $row_mem_overhead = $PERL_MEM_OVERHEAD * ($#{$self->{_in}->cols} + 2);
    
    $self->segment_start(\@rows);
    while ($fref = &$read_fastpath_sub) {
	push (@rows, $fref);
	$memory_used += $row_mem_overhead;
	foreach (@$fref) {
	    $memory_used += length($_);
	};
	if ($memory_used > $scaled_max_memory) {
	    $self->segment_overflow(\@rows);
	    $memory_used = 0;
	};
    };
    # handle end case
    $self->segment_overflow(\@rows, 1) if ($#rows > -1);   # (spill any records in queued)
    # merge, if necessary
    $self->segment_merge;
};
    


=head2 finish

    $filter->finish();

Internal: write trailer.

=cut
sub finish ($) {
    my($self) = @_;

    $self->SUPER::finish();
}


=head1 AUTHOR and COPYRIGHT

Copyright (C) 1991-2007 by John Heidemann <johnh@isi.edu>

This program is distributed under terms of the GNU general
public license, version 2.  See the file COPYING
with the distribution for details.

=cut

1;
